/*
 * NVIDIA CUDA Debugger CUDA-GDB Copyright (C) 2007-2015 NVIDIA Corporation
 * Written by CUDA-GDB team at NVIDIA <cudatools@nvidia.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "defs.h"
#include "inferior.h"
#include "gdbthread.h"
#include "arch-utils.h"
#include "regcache.h"

#include <block.h>

#include "cuda-autostep.h"
#include "cuda-state.h"
#include "cuda-iterator.h"
#include "cuda-frame.h"
#include "cuda-tdep.h"
#include "cuda-options.h"

/* When inside an autostep range, we go into single-step mode */
static bool autostep_stepping = false;
static bool autostep_pending = false;

/* From infcmd.c */
extern void delete_longjmp_breakpoint_cleanup (void *ignore);
extern void step_1 (int skip_subroutines, int single_inst, char *count_string);

/* Getters and setters */

bool
cuda_get_autostep_pending (void)
{
  return autostep_pending;
}

void
cuda_set_autostep_pending (bool pending)
{
  autostep_pending = pending;
}

bool
cuda_get_autostep_stepping (void)
{
  return autostep_stepping;
}

void
cuda_set_autostep_stepping (bool stepping)
{
  autostep_stepping = stepping;
}

bool
cuda_autostep_stop (void)
{
  struct thread_info *tp = NULL;

  if (target_has_execution && !ptid_equal (inferior_ptid, null_ptid))
    tp = inferior_thread ();

  return !tp || !tp->control.stop_step;
}

static void
autostep_report_exception_host (uint64_t before_pc)
{
  /* We know the exception must have been at the previous pc */
  struct gdbarch *gdbarch = get_current_arch ();
  struct type *type_uint32   = builtin_type (gdbarch)->builtin_uint32;
  struct type *type_data_ptr = builtin_type (gdbarch)->builtin_data_ptr;

  struct symtab_and_line before_sal = find_pc_line (before_pc, 0);

  printf_filtered (_("Autostep precisely caught exception at %s:%d (0x%llx)\n"),
    before_sal.symtab->filename, before_sal.line, (unsigned long long)before_pc);

  set_internalvar (lookup_internalvar ("autostep_exception_pc"),
    value_from_longest (type_data_ptr, (LONGEST) before_pc));
  set_internalvar (lookup_internalvar ("autostep_exception_line"),
    value_from_longest (type_uint32, (LONGEST) before_sal.line));
}

static void
autostep_report_exception_device (int nsteps, int before_ln, uint64_t before_pc, uint64_t after_pc)
{
  struct gdbarch *gdbarch    = get_current_arch ();
  struct type *type_uint32   = builtin_type (gdbarch)->builtin_uint32;
  struct type *type_data_ptr = builtin_type (gdbarch)->builtin_data_ptr;
  uint64_t exception_pc;
  struct symtab_and_line exception_sal;
  cuda_coords_t c;
  bool divergent;
  char exception_pc_line_info[200];

  gdb_assert (nsteps >= 1);

  cuda_print_message_focus (false);

  cuda_coords_get_current (&c);
  divergent = !lane_is_active (c.dev, c.sm, c.wp, before_ln);

  /* Calculate exception PC - if more than 1 one instruction was executed
   * it means there was no control flow instructions so after_pc should
   * be used as a reference (before_pc is far behind after_pc).
   * If just one instruction was executed use before_pc as a reference,
   * because the executed instruction could be a control flow instruction. */
  if (nsteps > 1 || divergent)
    cuda_api_get_adjusted_code_address (c.dev, after_pc, &exception_pc, CUDBG_ADJ_PREVIOUS_ADDRESS);
  else
    exception_pc = before_pc;

  exception_sal = find_pc_line (exception_pc, 0);

  if (exception_sal.symtab && exception_sal.line)
    snprintf (exception_pc_line_info, sizeof (exception_pc_line_info),
              "%s:%d (0x%llx)",
              exception_sal.symtab->filename, exception_sal.line,
              (unsigned long long)exception_pc);
  else
    snprintf (exception_pc_line_info, sizeof (exception_pc_line_info),
              "0x%llx",
              (unsigned long long)exception_pc);

  if (divergent)
    printf_filtered (_("Autostep caught exception at instruction before 0x%llx\n"
                       "This is probably %s\n"),
                     (unsigned long long)after_pc, exception_pc_line_info);
  else
    printf_filtered (_("Autostep precisely caught exception at %s\n"),
                     exception_pc_line_info);

  set_internalvar (lookup_internalvar ("autostep_exception_pc"),
                   value_from_longest (type_data_ptr, (LONGEST) exception_pc));
  set_internalvar (lookup_internalvar ("autostep_exception_line"),
                   value_from_longest (type_uint32, (LONGEST) exception_sal.line));
}

/* Single steps all the host code if it is at an autostep */
static void
handle_autostep_host (void)
{
  struct breakpoint *astep, *overlap;
  uint64_t before_pc, after_pc;
  int remaining;
  bool single_inst;

  /* Check we're at the an autostep bp */
  before_pc = stop_pc;
  astep = cuda_find_autostep_by_addr (before_pc);
  if (astep == NULL)
    {
      cuda_set_autostep_pending (false);
      return;
    }

  /* Check if we single step or step by lines */
  single_inst = astep->cuda_autostep_length_type == cuda_autostep_insts;

  /* This suppresses printing the line/instruction after each step */
  cuda_set_autostep_stepping (true);

  /* Step until we are out of the autostep range */
  remaining = astep->cuda_autostep_length;

  while (remaining > 0)
    {
      before_pc = regcache_read_pc (get_current_regcache ());

      /* Clear pending flag to test if we encounter another autostep */
      cuda_set_autostep_pending (false);

      /* Basically does a next/nexti */
      step_1 (false, single_inst, NULL);

      if (cuda_autostep_stop ())
        {
          struct thread_info *tp = NULL;

          if (target_has_execution && !ptid_equal (inferior_ptid, null_ptid))
            tp = inferior_thread ();

          if (tp && signal_pass_state (tp->suspend.stop_signal))
            {
              /* This is an exception */
              autostep_report_exception_host (before_pc);
            }
          else
            {
              /* This is a breakpoint or the user stopped the program. */
              warning (_("Program stopped during an autostep range.\n"
                "Autostepping will not resume upon continuing.\n"));
            }

          break;
        }

      after_pc = regcache_read_pc (get_current_regcache ());

      /* Handle overlapping autosteps */
      if (cuda_get_autostep_pending ())
        {
          overlap = cuda_find_autostep_by_addr (after_pc);
          warning (_("Overlapping autostep %d ignored"), overlap->number);
        }

      remaining--;
    }

  /* Mark that autostepping has been handled */
  cuda_set_autostep_pending (false);
  cuda_set_autostep_stepping (false);
}

static uint64_t
find_end_pc (uint64_t pc)
{
  struct block *bl;
  struct minimal_symbol *msymbol;

  bl = block_for_pc ((CORE_ADDR)pc);
  if (bl)
       return BLOCK_END (bl);

  msymbol = lookup_minimal_symbol_by_pc ((CORE_ADDR)pc);
  if (msymbol)
       return SYMBOL_VALUE_ADDRESS (msymbol) + MSYMBOL_SIZE (msymbol);

  return (uint64_t)-1LL;
}

static int
count_instructions (uint64_t pc, uint64_t end_pc)
{
  const char *inst;
  uint32_t inst_size;
  kernel_t kernel = cuda_current_kernel ();
  int count = 0;

  for (; pc < end_pc; pc += inst_size)
    {
      inst = kernel_disassemble (kernel, pc, &inst_size);
      if (!inst)
        break; /* Abort the loop if pc is outside of the routine boundary */
      if (inst[0] == 0)
        continue; /* Ignore empty instructions */
      ++count;
    }

  return count;
}

static int
count_lines (uint64_t pc, uint64_t end_pc, uint32_t inst_size)
{
  struct symtab_and_line cur_sal, next_sal;
  int nlines = 0;

  for (cur_sal = find_pc_line(pc, 0);
       pc <= end_pc;
       pc += inst_size, cur_sal = next_sal)
    {
      next_sal = find_pc_line(pc, 0);
      /* Check if line numbers differ.
       * If no line information exists treat each instruction as one line. */
      if (!(cur_sal.symtab && cur_sal.line) ||
          !(next_sal.symtab && next_sal.line) ||
          cur_sal.line != next_sal.line)
        ++nlines;
    }

  return nlines;
}

/* Single steps all the warps that are at an autostep */
static void
handle_autostep_device (void)
{
  cuda_iterator iter;
  cuda_coords_t after_coords;
  struct breakpoint *astep, *overlap;
  uint64_t before_pc, after_pc, end_pc;
  struct symtab_and_line before_sal, after_sal;
  int nsteps, lines, remaining;
  struct cleanup *old_cleanups;
  bool single_inst;
  cuda_coords_t filter;
  const char *sm_type;
  uint32_t inst_size;
  uint64_t warps_mask;

  /* This suppresses printing of the line after each step */
  cuda_set_autostep_stepping (true);

  /* Iterate through all warps in current grid that are at a breakpoint */

  filter = CUDA_WILDCARD_COORDS;
  filter.valid = true;
  filter.gridId = CUDA_CURRENT;
  cuda_coords_evaluate_current (&filter, false);

  iter = cuda_iterator_create (CUDA_ITERATOR_TYPE_THREADS, &filter,
    CUDA_SELECT_BKPT | CUDA_SELECT_VALID);
  old_cleanups = make_cleanup ((make_cleanup_ftype*)cuda_iterator_destroy,
    (void*)iter);

  cuda_iterator_start (iter);
  while (cuda_focus_is_device () && !cuda_iterator_end (iter))
    {
      cuda_coords_t curc, nextc, c;
      int before_ln;

      /* We look only at logical coordinates from iterator.
         Physical coordinates can change with Software Preemption enabled
         after each step (on all warps, not just currently stepped). */
      curc = cuda_iterator_get_current (iter);
      c = curc; /* This will be used as an up to date physical coordinate holder. */

      /* Update physical coordinates */
      cuda_coords_complete_physical (&c);

      /* Make sure the warp didn't become invalid */
      if (!warp_is_valid (c.dev, c.sm, c.wp))
        goto next_warp;

      /* Check we're at the autostep bp */
      before_pc = warp_get_active_virtual_pc (c.dev, c.sm, c.wp);
      astep = cuda_find_autostep_by_addr (before_pc);
      if (astep == NULL || astep->enable_state != bp_enabled)
        goto next_warp;

      /* Check that the device is Fermi or better */
      /* Must check here in case user re-enabled it */
      sm_type = device_get_sm_type (c.dev);
      if (strncmp (sm_type, "sm_1", 4) == 0)
        {
          warning ("Disabling autostep %d on device %d because autostep "
            "requires compute capability 2.0 or higher.", astep->number, c.dev);
          astep->enable_state = bp_disabled;
          goto next_warp;
        }

      /* Check if we single step or step by lines */
      single_inst = astep->cuda_autostep_length_type == cuda_autostep_insts;

      /* Set focus to current warp */
      before_ln = warp_get_lowest_active_lane (c.dev, c.sm, c.wp);
      cuda_coords_set_current_physical (c.dev, c.sm, c.wp, before_ln);

      /* Step until we are out of the autostep range */
      remaining = astep->cuda_autostep_length;

      while (remaining > 0)
        {
          /* Clear pending flag to test if we encounter another autostep */
          cuda_set_autostep_pending (false);

          before_pc = warp_get_active_virtual_pc (c.dev, c.sm, c.wp);
          before_sal = find_pc_line(before_pc, 0);

          end_pc = -1; /* Not limited at the beginning */

          /* Limit end_pc; at first assume there are no control flow instructions */
          if (!cuda_options_single_stepping_optimizations_enabled ())
            {
              end_pc = before_pc;
            }
          else if (!(before_sal.symtab && before_sal.line) || single_inst)
            {
              /* Get instruction size */
              kernel_disassemble (cuda_current_kernel (), before_pc, &inst_size);

              end_pc = before_pc + remaining * inst_size;
            }
          else
            {
              VEC (CORE_ADDR) *line_pcs;
              CORE_ADDR line_pc;
              int i;
              struct linetable_entry *best_item = NULL; /* ignored */

              /* Search for all PCs which correspond to (current + remaining) line.
               * Try to pick up lowest possible address after current pc
               * corresponding to limiting line. */

              line_pcs = find_pcs_for_symtab_line (before_sal.symtab,
                                                   before_sal.line + remaining,
                                                   &best_item);

              for (i = 0; VEC_iterate (CORE_ADDR, line_pcs, i, line_pc); ++i)
                {
                  if (before_pc < line_pc && line_pc < end_pc)
                    end_pc = line_pc;
                }

              VEC_free (CORE_ADDR, line_pcs);
            }

          /* If pc is in the top frame - do not allow autostepping outside of kernel boundaries */
          if (cuda_frame_outermost_p (get_next_frame (get_current_frame ())))
            {
              uint64_t kernel_end_pc;
              kernel_end_pc = find_end_pc (before_pc);
              if (before_pc >= kernel_end_pc)
                break;
              if (kernel_end_pc < end_pc)
                end_pc = kernel_end_pc;
            }

          /* Calculate how many steps should be taken */
          if (cuda_options_single_stepping_optimizations_enabled ())
            end_pc = cuda_find_next_control_flow_instruction (before_pc, end_pc, false, &inst_size);
          if (end_pc == before_pc)
            nsteps = 1; /* Currently at control flow instruction */
          else
            nsteps = count_instructions (before_pc, end_pc);

          /* Does stepi, but cuda_sstep_execute takes cuda_sstep_nsteps into
           * account to execute 'single step nsteps times' */
          cuda_sstep_set_nsteps(nsteps);
          step_1 (false, true, NULL);

          /* Check if logical coordinates are still valid and update physical
             coordinates. If logical coordinates are not valid, warp ran to
             completion. */
          if (cuda_coords_complete_physical (&c))
            break;

          /* Make sure we can continue stepping this warp */
          if (!cuda_focus_is_device () || !warp_is_valid (c.dev, c.sm, c.wp))
            break;

          after_pc = warp_get_active_virtual_pc (c.dev, c.sm, c.wp);
          after_sal = find_pc_line(after_pc, 0);

          cuda_trace_domain (CUDA_TRACE_BREAKPOINT,
                             "Autostep: issued single step %d steps (from %llx to %llx).",
                             nsteps, before_pc, end_pc);

          if (cuda_autostep_stop ())
            {
              struct thread_info *tp = NULL;

              if (target_has_execution && !ptid_equal (inferior_ptid, null_ptid))
                tp = inferior_thread ();

              if (tp && signal_pass_state (tp->suspend.stop_signal))
                {
                  /* This is an exception */
                  autostep_report_exception_device (nsteps, before_ln, before_pc, after_pc);
                  cuda_set_autostep_pending (false);
                }
              else
                {
                  /* This is a breakpoint or the user stopped the program. */
                  warning (_("Program stopped during an autostep range.\n"
                   "This warp will not autostep upon resuming."));
                  /* Resume any pending warps but not current warp. */
                  cuda_set_autostep_pending (true);
                }

              cuda_set_autostep_stepping (false);

              do_cleanups (old_cleanups);
              return;
            }

          /* Find out how many lines/nsteps were actually stepped */
          if (nsteps > 1)
            {
              nsteps = count_instructions (before_pc, after_pc);
              lines = count_lines (before_pc, after_pc, inst_size);
            }
          else /* Control flow instruction */
            {
              gdb_assert(nsteps == 1);

              /* Calculate lines - if no line information exists treat it as one instruction */
              lines = !(before_sal.symtab && before_sal.line) ||
                      !(after_sal.symtab && after_sal.line) ||
                      before_sal.line != after_sal.line;
            }

          remaining -= single_inst ? nsteps : lines;

          cuda_trace_domain (CUDA_TRACE_BREAKPOINT,
                             "Autostep: in fact single stepped %d steps / %d lines (%d %s left). "
                             "PC after is %llx (%d).",
                             nsteps, lines, remaining, single_inst ? "instructions" : "lines",
                             (unsigned long long)after_pc, after_sal.line);

          /* Handle overlapping autosteps */
          if (cuda_get_autostep_pending ())
            {
              overlap = cuda_find_autostep_by_addr (after_pc);
              if (remaining > 0)
                {
                  warning (_("Overlapping autostep %d ignored"), overlap->number);
                }
              else
                {
                  /* The last astep just finished, go onto the next one */
                  astep = overlap;
                  remaining = overlap->cuda_autostep_length;
                }
            }
        }

next_warp:
      cuda_trace_domain (CUDA_TRACE_BREAKPOINT,
                         "Autostep: handling next warp! Previous was: tId=(%d,%d,%d) bId=(%d,%d,%d)",
                         curc.threadIdx.x, curc.threadIdx.y, curc.threadIdx.y,
                         curc.blockIdx.x, curc.blockIdx.y, curc.blockIdx.z);

      /* Skip to next warp (by using possibly outdated physical coordinates,
         but sorted correctly by logical coordinates) */
      do
        {
        cuda_iterator_next (iter);
        nextc = cuda_iterator_get_current (iter);
        }
      while (cuda_focus_is_device () && !cuda_iterator_end (iter) &&
             curc.dev == nextc.dev && curc.sm == nextc.sm && curc.wp == nextc.wp);
    }

  /* Mark that autostepping has been handled */
  cuda_set_autostep_pending (false);
  cuda_set_autostep_stepping (false);

  do_cleanups (old_cleanups);
}

void
cuda_handle_autostep (void)
{
  if (!cuda_get_autostep_pending ())
    return;

  if (cuda_focus_is_device ())
    handle_autostep_device ();
  else
    handle_autostep_host ();
}

